static const char rcsid[] = 
"$Id: display.c,v 1.4 1996/06/20 18:58:16 schwartz Exp $";

#include "globin/genie.h"
#include "globin/genielib.h"
#include "globin/errlib.h"
#include "globin/utillib.h"
                            
#include <stdio.h>          
#include <stdlib.h>         
#include <ctype.h>          
#include <stdarg.h>         
#include <string.h>

#define NIL (char*)0

STATIC void show_attr (int level, attribute *attr);
STATIC void show_mod (int level, modifier *mod);
STATIC int show_inst (int level, instance *inst, char* delim);
STATIC void display (int level, char *item1, ... );
STATIC int show_tuplist (int level, int type, tuplink* tuplist);

void show_var (int level, varlink *var);
int show_tup (int level, tuplink *tup, char* delim);

void show_var (int level, varlink *var) 
{
    int i;

    if (!var)
	return;

    if (var->tag) {
	display (level, "<", var->tag, ">  =>\n", NIL);
	level++;
    }
    for (i = 0; i < var->numattr; i++)
	show_attr (level, var->attrs + i);
    if (i == 0) {
	display (level, "EMPTY\n", NIL);
    }
    if (var->attrs[i].type != MARK)
	genie_fatal ("show_var", "bad attribute count", NIL);
}

struct attributes {
  char* s_left;
  char* s_right;
  char* t_left;
  char* t_right;
  int level;
};

struct attributes attributes[] = {
    [INT]  = { "integer",     "",  "",       "",  0 },
    [REAL] = { "real",        "",  "",       "",  0 },
    [STR]  = { "string",      "",  "",       "",  0 },
    [BOOL] = { 0 },

/*
    [REC]  = { "RECORD (\n",  ")", "(",      ")", 1 },
    [SET]  = { "SET_OF {\n",  "}", "set([",  "])", 1 },
    [LIST] = { "LIST_OF {\n", "}", "list([", "])", 1 },
    [VAR]  = { "VARIANT [\n", "]", "var(",   "))", 1 },
Gudjon
*/

    [REC]  = { "RECORD (\n",  ")", "",      "", 1 },
    [SET]  = { "SET_OF {\n",  "}", "({[",  "]})", 1 },
    [LIST] = { "LIST_OF {\n", "}", "([", "])", 1 },
    [VAR]  = { "VARIANT [\n", "]", "(+",   ")", 1 },
    [UNK]  = { 0 },
    [ERR_TYPE] = { 0 },
    [MARK] = { 0 },
    { 0 },
};

STATIC void show_attr (int level, attribute *attr) 
{
    varlink *vptr;		/* ptr to variant              */
    modifier *mptr;		/* ptr to modifier             */
    type_form type;

    if (!attr)
	return;

    display (level, attr->name, ": ", NIL);

    type = attr->type;
    if (type > MARK)
	genie_fatal ("show_attr", "bad attribute type", NIL);	 

    display (level, attributes[type].s_left, NIL);
    for (vptr = attr->varlist; vptr; next (vptr))
	show_var (level + attributes[type].level, vptr);
    display (level, attributes[type].s_right, NIL);

    for (mptr = attr->modlist; mptr; next (mptr))
	show_mod (level, mptr);

/*
    if (is_composite (attr->type))
	display (level, "  #end ", attr->name, NIL);
*/

    display (0, "\n", NIL);
    if (!level) {
	display(0, "\n", NIL);	/* blank line between top-level attributes */
    }
}


STATIC void show_mod (int level, modifier *mod) 
{
    char **cptr;		/* ptr to enum choice          */

    if (!mod)
	return;

    switch (mod->form) {
    case OPT:
	display (level, " OPTIONAL", NIL);
	break;
    case KEY:
	display (level, " KEY", NIL);
	break;
    case FKEY:
	display (level, " FOREIGN_KEY ", mod->u.tablename, NIL);
	break;
    case ENUM:
	display (level, " CHOICE_OF [", NIL);
	for (cptr = mod->u.choices ; *cptr; cptr++)
	    display (level, cptr[0], (cptr[1] ? ", " : ""), NIL);
	display (level, "]", NIL);
	break;
    case REXP:
	display (level, " REGEXP '", mod->u.regexp, "'", NIL);
	break;
    case RSTR:
	display (level, " RESTRICT <not implemented>", NIL);
	break;
    default:
	genie_fatal ("show_mod", "bad modifier form", NIL);
    }

} 

int show_tup (int level, tuplink *tup, char* delim)
{
    instance *iptr; 
    char buf[128];

    if (!tup)
	return 0;

    for (iptr = tup->values; iptr && iptr->attrptr; iptr++) {
	show_inst (level, iptr, buf);
	if (iptr[1].attrptr)
	    display (level, delim, NIL);
    }

    return (iptr - tup->values);
}

STATIC int show_inst (int level, instance *inst, char* delim) 
{
    attribute *aptr;		/* ptr to schema attribute     */
    char *sval;			/* string version of value     */
    int n = 0;

    if (!inst)
	return 0;
    aptr = inst->attrptr;
    if (!aptr)
	return 0;

    if (is_composite (aptr->type)) {
	display (level, aptr->name, "", NIL);
	display (level, attributes[aptr->type].t_left, "", NIL);
	if (aptr->type == VAR)
	    display(level+1, inst->tag, "", NIL);
        show_tuplist(level+1, aptr->type, inst->u.t.tuplist);
	display (level+1, attributes[aptr->type].t_right, "\n", NIL);
	display (level, "\n", NIL);
    } else {
	n = 1;
	sval = stringval (inst);
	display (level, aptr->name, "(", sval, ")", NIL);
	ckfree (sval);
    }

    return n;
}

STATIC int list_length(tuplink* tptr)
{
    int n;
    for (n = 0; tptr; next (tptr))
	++n;
    return n;
}

STATIC int show_tuplist (int level, int type, tuplink* tuplist)
{
    tuplink* tptr;

    for (tptr = tuplist; tptr; next (tptr)) {
	display (level+1, "(", NIL);
	if (show_tup (level+2, tptr, ",\n") == 0)
	    display (level+2, "null", NIL);
	display (level+1, ")%tptr\n", NIL);
	if (tptr->next)
	    display (level, ", %\n", NIL);
    }
    return list_length(tuplist);
}

static int curcol = 0;

STATIC void cputs (char* item)
{
    char* p = strrchr(item, '\n');
    if (p) curcol = 0;
    curcol += strlen(p ? p+1 : item);
    fputs(item, stdout);
}

STATIC void display (int level, char *item1, ... ) {
    va_list ap;
    char *item;
    
    if (curcol == 0) {
	while (level--)
	    putchar(' ');
    }
    va_start (ap, item1);
    for (item = item1; item; item = va_arg (ap, char *))
	cputs(item);
    va_end (ap);
}

