user:portray_message(_,_).

:- use_module(library(lists)).

report(Term) :- report(Term,0), nl.

report(null,_) :- !.
report(Term,Ind) :- functor(Term,Name,Arity), report(Name,Arity,Term,Ind).

report(Name,1,Term,Ind) :-
  arg(1,Term,Arg), integer(Arg), !, format('~n~*|~a = ~d',[Ind,Name,Arg]).
report(Name,1,Term,Ind) :-
  arg(1,Term,Arg), float(Arg), !, format('~n~*|~a = ~5f',[Ind,Name,Arg]).
report(Name,1,Term,Ind) :-
  arg(1,Term,Arg), atom(Arg), Arg\==[], Arg\==null, !,
  format('~n~*|~a = ~q',[Ind,Name,Arg]).
report(Name,1,Term,Ind) :-
  arg(1,Term,null), !, format('~n~*|~a = NULL',[Ind,Name]).
report(Name,1,Term,Ind) :-
  arg(1,Term,List), is_list(List), !, report_list(Name,List,Ind).
report(Name,1,Term,Ind) :-
  arg(1,Term,{List}), is_list(List), !, report_list(Name,List,Ind).
report(Name,1,Term,Ind) :-
  arg(1,Term,+Var), !, functor(Var,VarName,VarArity),
  format('~n~*|BEGIN ~a AS ~a',[Ind,Name,VarName]),
  NewInd is Ind+2, report_record(1,VarArity,Var,NewInd),
  format('~n~*|END ~a',[Ind,Name]).
report(Name,Arity,Term,Ind) :-
  format('~n~*|BEGIN ~a',[Ind,Name]),
  NewInd is Ind+2, report_record(1,Arity,Term,NewInd),
  format('~n~*|END ~a',[Ind,Name]).

report_list(_,[],_).
report_list(Name,[First|Rest],Ind) :-
  format('~n~*|BEGIN ~a',[Ind,Name]),
  NewInd is Ind+2, report_anon_rec(First,NewInd),
  format('~n~*|END ~a',[Ind,Name]),
  report_list(Name,Rest,Ind).

report_anon_rec((First,Rest),Ind) :- !,
  report(First,Ind), report_anon_rec(Rest,Ind).
report_anon_rec(Last,Ind) :- report(Last,Ind).

report_record(ArgNum,Arity,_,_) :- ArgNum>Arity, !.
report_record(ArgNum,Arity,Term,Ind) :-
  arg(ArgNum,Term,Arg), report(Arg,Ind),
  NewArgNum is ArgNum+1, report_record(NewArgNum,Arity,Term,Ind).

contains_match(LongString, ShortString) :-
        atom_chars(LongString, LongCh),
        atom_chars(ShortString, ShortCh),
        contains_match_chars(LongCh, ShortCh).

% succeeds if the first list of characters contains the second list, in order
contains_match_chars(_, []).
contains_match_chars([], _) :- !, fail.
contains_match_chars([LongCh|RemLongChars], [ShortCh|RemShortCh]) :-
        /* if first character matches, find remaining adjacent characters */
        chars_match([LongCh], [ShortCh]),
        contains_rest_match_chars(RemLongChars, RemShortCh), !.
contains_match_chars([_|LongChars], ShortCh) :-
        contains_match_chars(LongChars, ShortCh), !.
chars_match(Ch, Ch).
chars_match("C", "B").
chars_match("G", "B").
chars_match("T", "B").
chars_match("A", "D").
chars_match("G", "D").
chars_match("T", "D").
chars_match("A", "H").
chars_match("C", "H").
chars_match("T", "H").
chars_match("G", "K").
chars_match("T", "K").
chars_match("A", "M").
chars_match("C", "M").
chars_match("A", "N").
chars_match("C", "N").
chars_match("G", "N").
chars_match("T", "N").
chars_match("A", "R").
chars_match("G", "R").
chars_match("C", "S").
chars_match("G", "S").
chars_match("A", "V").
chars_match("C", "V").
chars_match("G", "V").
chars_match("A", "W").
chars_match("T", "W").
chars_match("A", "X").
chars_match("C", "X").
chars_match("T", "X").
chars_match("C", "Y").
chars_match("T", "Y").

% succeeds if the occurence of the second list of characters has been
%     found at the beginning of the first list
contains_rest_match_chars(_, []).
contains_rest_match_chars([LongCh|RemLongChars], [ShortCh|RemShortCh]) :-
        /* if first character matches, find remaining adjacent characters */
        chars_match([LongCh], [ShortCh]),
        contains_rest_match_chars(RemLongChars, RemShortCh), !.

overlaps(Region1,Region2) :-
  decode_region(Region1,Origin1,Start1,Stop1),
  decode_region(Region2,Origin2,Start2,Stop2),
  overlaps(Origin1,Start1,Stop1,Origin2,Start2,Stop2).

construct_gap(N1,N2,ConstructSegment,Size) :-
  findall(
    SegmentSize,
    (
      nth(N,ConstructSegment,Segment),
      N1<N, N<N2, get_segment_size(Segment,SegmentSize)
    ),
    SizeList
  ),
  sum_list(SizeList,Size).

get_segment_size(segment(segment_element(+beta_g_region(region(_,
	                                                       start(Start),
							       stop(Stop)),
						        _,_)),
			 _),
	         Size) :-
  Size is Stop-Start+1.
get_segment_size(segment(segment_element(+non_beta_g_region(_,_,_,
	                                                    length_basepairs(
							      Size)),
					 _)),
		 Size).
get_segment_size(segment(segment_element(+oligo(sequence(Atom)),_)),Size) :-
  atom_chars(Atom,String), length(String,Size).

%%%

decode_region(region(origin(Origin),start(Start),stop(Stop)),
	      Origin,Start,Stop).
decode_region(nucleotide(origin(Origin),start(Start),stop(Stop)),
	      Origin,Start,Stop).
decode_region(abs_range(origin(Origin),start(Start),stop(Stop)),
	      Origin,Start,Stop).

apply_offset(0,_,_) :- abort.
apply_offset(X,D,Y) :-
	T is X + D, adjust_offset(X,T,Y).

adjust_offset(X,T,Y) :-
	X < 0, T >= 0, !, Y is T + 1.
adjust_offset(X,T,Y) :-
	X > 0, T =< 0, !, Y is T - 1.
adjust_offset(_,T,T).

resolve_origin(Origin,Start1,Stop1, Species,Start2,Stop2) :-
	origin_info({OI}),
	memberchk((name(Origin),species(_),file(Species),offset(Offset)),OI),
	!,
	apply_offset(Start1, Offset - 1, Start2),
	apply_offset(Stop1, Offset - 1, Stop2).
resolve_origin(Species,Start,Stop, Species,Start,Stop).

%%%   Sum the intervals in an alignment up to a given offset in the sequence, 
%%% returning the offset in the alignment. :-
%%%   If the list is empty, return the current alignment offset.
%%%   Otherwise, if the current sequence offsetis bigger than the 
%%% offset we are looking for, return the current alignment offset.
%%%   Otherwise, if the element is a gap, increment the alignment offset
%%% by the size of the gap and continue.
%%%   Otherwise, if the element is a match, increment the sequence offset
%%% by the size of the gap and continue.
sump(_,[],_,Aoff,Aoff).
sump(Seqoff,_,Soff,Aoff,Aoff) :- Soff >= Seqoff.
sump(Seqoff,[Gap|Rest],Soff,Aoff0,Aoff1) :- Soff < Seqoff, Gap < 0,
	A is Aoff0 - Gap, 
	sump(Seqoff,Rest,Soff,A,Aoff1).
sump(Seqoff,[Seg|Rest],Soff,Aoff0,Aoff1) :- Soff < Seqoff, Seg >= 0,
	S is Soff + Seg,
	%%A is Aoff0 + Seg,
	sump(Seqoff,Rest,S,Aoff0,Aoff1).

%%% Handle offsets for sump
osump(Seqoff,Seqorg,[Seg|Rest],Soff,Aoff,Aoffout) :-
	X is Seqoff-Seqorg+1,
	sump(X,[Seg|Rest],Soff,Aoff,T),
	Aoffout is T + Seqoff - Seqorg.

seosump(Startoff/Endoff,Seqorg,[Seg|Rest],Soff,Aoff,StartAoffout/EndAoffout) :-
	osump(Startoff,Seqorg,[Seg|Rest],Soff,Aoff,StartAoffout),
	osump(Endoff,Seqorg,[Seg|Rest],Soff,Aoff,EndAoffout).


seqpos_to_alignpos(Seqno,Seqoff,Alignpos) :-
        malign(MA),
	nth(Seqno, MA, align(Species,Start,End,Gaplist)),
	sump(Seqoff-Start+1,Gaplist,0,0,A),
	%%format("Species ~w (~w,~w) -> ~w~n",[Species,Start,End,A]),
	Alignpos is A+Seqoff-Start.

overlaps(Origin1,Start1,Stop1,Origin2,Start2,Stop2) :-
	resolve_origin(Origin1,Start1,Stop1, Species1,Start11,Stop12),
	resolve_origin(Origin2,Start2,Stop2, Species2,Start21,Stop22),
	malign(MA),

	%%format('~w ~w ~w//~w ~w ~w~n',[Origin1,Start1,Stop1,Species1,Start11,Stop12]),
	%%format('~w ~w ~w//~w ~w ~w~n',[Origin2,Start2,Stop2,Species2,Start21,Stop22]),

	member(align(Species1,Org1,Eee,Gaplist1), MA),
	osump(Start11,Org1,Gaplist1,0,0,S1),
	osump(Stop12,Org1,Gaplist1,0,0,E1),

	%%format('~w S1=~w T1=~w Start11=~w Org1=~w ~w~n',[Species1,S1,T1,Start11,Org1,Eee]),
	%%format('~w E1=~w T2=~w Stop12=~w  Org1=~w ~w~n',[Species1,E1,T2,Stop12,Org1,Eee]),

	member(align(Species2,Org2,_,Gaplist2), MA),
	osump(Start21,Org2,Gaplist2,0,0,S2),
	osump(Stop22,Org2,Gaplist2,0,0,E2),

	%%format('~w S2=~w T3=~w  Start21=~w Org2=~w~n', [Species2,S2,T3,Start21,Org2]),
	%%format('~w E2=~w T4=~w  Stop22=~w  Org2=~w~n', [Species2,E2,T4,Stop22,Org2]),

	%%format('overlap ~w,~w X ~w,~w~n',[S1,E1,S2,E2]),
	S1 =< E2, S2 =< E1,
	!.

contains_region(Region1,Region2) :-
  decode_region(Region1,Origin1,Start1,Stop1),
  decode_region(Region2,Origin2,Start2,Stop2),
  contains_region(Origin1,Start1,Stop1,Origin2,Start2,Stop2).

contains_region(Origin1,Start1,Stop1,Origin2,Start2,Stop2) :-
	resolve_origin(Origin1,Start1,Stop1, Species1,Start11,Stop12),
	resolve_origin(Origin2,Start2,Stop2, Species2,Start21,Stop22),
	malign(MA),

	member(align(Species1,Org1,_,Gaplist1), MA),
	osump(Start11,Org1,Gaplist1,0,0,S1),
	osump(Stop12,Org1,Gaplist1,0,0,E1),

	member(align(Species2,Org2,_,Gaplist2), MA),
	osump(Start21,Org2,Gaplist2,0,0,S2),
	osump(Stop22,Org2,Gaplist2,0,0,E2),

	S1=<S2, E1>=E2,
	!.

:-
  compile('/depot/globin/data/yama2'), compile('/depot/globin/data/tuples'),
  save(glop,Res), (Res==1 -> prompt(_,''), [user], halt; halt).
